# Generated by Django 4.1.9 on 2023-07-11 08:24

import django.db.models.deletion
import wbfdm.models.fk_fields
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbfdm", "0002_rename_statements_instrumentlookup_financials_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="instrument",
            name="estimate_backend",
            field=wbfdm.models.fk_fields.CompositeForeignKey(
                null=True,
                null_if_equal=[],
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="wbfdm.instrumentlookup",
                to_fields={
                    "estimates": wbfdm.models.fk_fields.RawFieldValue(True),
                    "instrument_id": wbfdm.models.fk_fields.LocalFieldValue("id"),
                },
            ),
        ),
        migrations.AlterField(
            model_name="instrumentlookup",
            name="parameters",
            field=models.JSONField(blank=True, default=dict, null=True),
        ),
    ]
