# Generated by Django 5.0.9 on 2024-11-06 13:20

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("metric", "0001_initial"),
        ("wbfdm", "0027_remove_instrument_unique_ric_and_more"),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name="instrumentmetric",
            name="unique_instrument_metric",
        ),
        migrations.AddConstraint(
            model_name="instrumentmetric",
            constraint=models.UniqueConstraint(
                fields=("basket_content_type", "basket_id", "instrument", "date", "key"),
                name="unique_instrument_metric",
                nulls_distinct=False,
            ),
        ),
    ]
