# Generated by Django 5.0.6 on 2024-07-04 07:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("currency", "0001_initial"),
        ("geography", "0001_initial"),
        ("io", "0007_alter_exportsource_query_params"),
        ("tags", "0001_initial"),
        ("wbfdm", "0022_instrument_cusip_option_open_interest_20d_and_more"),
    ]

    operations = [
        migrations.AddConstraint(
            model_name="instrument",
            constraint=models.UniqueConstraint(
                condition=models.Q(("is_security", True)), fields=("refinitiv_identifier_code",), name="unique_ric"
            ),
        ),
        migrations.AddConstraint(
            model_name="instrument",
            constraint=models.UniqueConstraint(
                condition=models.Q(("is_security", True)), fields=("refinitiv_mnemonic_code",), name="unique_rmc"
            ),
        ),
        migrations.AddConstraint(
            model_name="instrument",
            constraint=models.UniqueConstraint(
                condition=models.Q(("is_security", True)), fields=("isin",), name="unique_isin"
            ),
        ),
        migrations.AddConstraint(
            model_name="instrument",
            constraint=models.UniqueConstraint(
                condition=models.Q(("is_security", True)), fields=("sedol",), name="unique_sedol"
            ),
        ),
        migrations.AddConstraint(
            model_name="instrument",
            constraint=models.UniqueConstraint(
                condition=models.Q(("is_security", True)), fields=("valoren",), name="unique_valoren"
            ),
        ),
        migrations.AddConstraint(
            model_name="instrument",
            constraint=models.UniqueConstraint(
                condition=models.Q(("is_security", True)), fields=("cusip",), name="unique_cusip"
            ),
        ),
    ]
