class MetricInvalidParameterException(Exception):
    """
    If instantiated backend parameters are invalid, the orchestrator or backend itself are expected to raise that exception
    """

    pass
